#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

from muninn._compat import string_types as basestring

import copy
import datetime
import re
import uuid

import muninn.geometry as geometry

from muninn.enum import Enum
from muninn.exceptions import Error, InternalError
from muninn.function import Prototype, FunctionTable
from muninn.schema import Namespace, Boolean, Geometry, Integer, Long, Real, Sequence, Text, Timestamp, UUID
from muninn.visitor import Visitor


#
# Table of all supported operators and functions
#
type_map = {
    UUID: Boolean,
}
function_table = FunctionTable(type_map=type_map)
#
# Logical operators
#
function_table.add(Prototype("not", (Boolean,), Boolean))
function_table.add(Prototype("and", (Boolean, Boolean), Boolean))
function_table.add(Prototype("or", (Boolean, Boolean), Boolean))
#
# Membership operators
#
function_table.add(Prototype("in", (Integer, Sequence), Boolean))
function_table.add(Prototype("in", (Long, Sequence), Boolean))
function_table.add(Prototype("in", (Real, Sequence), Boolean))
function_table.add(Prototype("in", (Text, Sequence), Boolean))
function_table.add(Prototype("not in", (Integer, Sequence), Boolean))
function_table.add(Prototype("not in", (Long, Sequence), Boolean))
function_table.add(Prototype("not in", (Real, Sequence), Boolean))
function_table.add(Prototype("not in", (Text, Sequence), Boolean))
#
# Comparison operators
#
function_table.add(Prototype("==", (Long, Long), Boolean))
function_table.add(Prototype("==", (Long, Integer), Boolean))
function_table.add(Prototype("==", (Integer, Long), Boolean))
function_table.add(Prototype("==", (Integer, Integer), Boolean))
function_table.add(Prototype("==", (Real, Real), Boolean))
function_table.add(Prototype("==", (Real, Long), Boolean))
function_table.add(Prototype("==", (Long, Real), Boolean))
function_table.add(Prototype("==", (Real, Integer), Boolean))
function_table.add(Prototype("==", (Integer, Real), Boolean))
function_table.add(Prototype("==", (Boolean, Boolean), Boolean))
function_table.add(Prototype("==", (Text, Text), Boolean))
function_table.add(Prototype("==", (Timestamp, Timestamp), Boolean))
function_table.add(Prototype("==", (UUID, UUID), Boolean))

function_table.add(Prototype("!=", (Long, Long), Boolean))
function_table.add(Prototype("!=", (Long, Integer), Boolean))
function_table.add(Prototype("!=", (Integer, Long), Boolean))
function_table.add(Prototype("!=", (Integer, Integer), Boolean))
function_table.add(Prototype("!=", (Real, Real), Boolean))
function_table.add(Prototype("!=", (Real, Long), Boolean))
function_table.add(Prototype("!=", (Long, Real), Boolean))
function_table.add(Prototype("!=", (Real, Integer), Boolean))
function_table.add(Prototype("!=", (Integer, Real), Boolean))
function_table.add(Prototype("!=", (Boolean, Boolean), Boolean))
function_table.add(Prototype("!=", (Text, Text), Boolean))
function_table.add(Prototype("!=", (Timestamp, Timestamp), Boolean))
function_table.add(Prototype("!=", (UUID, UUID), Boolean))

function_table.add(Prototype("<", (Long, Long), Boolean))
function_table.add(Prototype("<", (Long, Integer), Boolean))
function_table.add(Prototype("<", (Integer, Long), Boolean))
function_table.add(Prototype("<", (Integer, Integer), Boolean))
function_table.add(Prototype("<", (Real, Real), Boolean))
function_table.add(Prototype("<", (Real, Long), Boolean))
function_table.add(Prototype("<", (Long, Real), Boolean))
function_table.add(Prototype("<", (Real, Integer), Boolean))
function_table.add(Prototype("<", (Integer, Real), Boolean))
function_table.add(Prototype("<", (Text, Text), Boolean))
function_table.add(Prototype("<", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype(">", (Long, Long), Boolean))
function_table.add(Prototype(">", (Long, Integer), Boolean))
function_table.add(Prototype(">", (Integer, Long), Boolean))
function_table.add(Prototype(">", (Integer, Integer), Boolean))
function_table.add(Prototype(">", (Real, Real), Boolean))
function_table.add(Prototype(">", (Real, Long), Boolean))
function_table.add(Prototype(">", (Long, Real), Boolean))
function_table.add(Prototype(">", (Real, Integer), Boolean))
function_table.add(Prototype(">", (Integer, Real), Boolean))
function_table.add(Prototype(">", (Text, Text), Boolean))
function_table.add(Prototype(">", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype("<=", (Long, Long), Boolean))
function_table.add(Prototype("<=", (Long, Integer), Boolean))
function_table.add(Prototype("<=", (Integer, Long), Boolean))
function_table.add(Prototype("<=", (Integer, Integer), Boolean))
function_table.add(Prototype("<=", (Real, Real), Boolean))
function_table.add(Prototype("<=", (Real, Long), Boolean))
function_table.add(Prototype("<=", (Long, Real), Boolean))
function_table.add(Prototype("<=", (Real, Integer), Boolean))
function_table.add(Prototype("<=", (Integer, Real), Boolean))
function_table.add(Prototype("<=", (Text, Text), Boolean))
function_table.add(Prototype("<=", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype(">=", (Long, Long), Boolean))
function_table.add(Prototype(">=", (Long, Integer), Boolean))
function_table.add(Prototype(">=", (Integer, Long), Boolean))
function_table.add(Prototype(">=", (Integer, Integer), Boolean))
function_table.add(Prototype(">=", (Real, Real), Boolean))
function_table.add(Prototype(">=", (Real, Long), Boolean))
function_table.add(Prototype(">=", (Long, Real), Boolean))
function_table.add(Prototype(">=", (Real, Integer), Boolean))
function_table.add(Prototype(">=", (Integer, Real), Boolean))
function_table.add(Prototype(">=", (Text, Text), Boolean))
function_table.add(Prototype(">=", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype("~=", (Text, Text), Boolean))

function_table.add(Prototype("+", (Long,), Long))
function_table.add(Prototype("+", (Integer,), Integer))
function_table.add(Prototype("+", (Real,), Real))

function_table.add(Prototype("-", (Long,), Long))
function_table.add(Prototype("-", (Integer,), Integer))
function_table.add(Prototype("-", (Real,), Real))

function_table.add(Prototype("+", (Long, Long), Long))
function_table.add(Prototype("+", (Long, Integer), Long))
function_table.add(Prototype("+", (Integer, Long), Long))
function_table.add(Prototype("+", (Integer, Integer), Integer))
function_table.add(Prototype("+", (Real, Real), Real))
function_table.add(Prototype("+", (Real, Long), Real))
function_table.add(Prototype("+", (Long, Real), Real))
function_table.add(Prototype("+", (Real, Integer), Real))
function_table.add(Prototype("+", (Integer, Real), Real))

function_table.add(Prototype("-", (Long, Long), Long))
function_table.add(Prototype("-", (Long, Integer), Long))
function_table.add(Prototype("-", (Integer, Long), Long))
function_table.add(Prototype("-", (Integer, Integer), Integer))
function_table.add(Prototype("-", (Real, Real), Real))
function_table.add(Prototype("-", (Real, Long), Real))
function_table.add(Prototype("-", (Long, Real), Real))
function_table.add(Prototype("-", (Real, Integer), Real))
function_table.add(Prototype("-", (Integer, Real), Real))

function_table.add(Prototype("*", (Long, Long), Long))
function_table.add(Prototype("*", (Long, Integer), Long))
function_table.add(Prototype("*", (Integer, Long), Long))
function_table.add(Prototype("*", (Integer, Integer), Integer))
function_table.add(Prototype("*", (Real, Real), Real))
function_table.add(Prototype("*", (Real, Long), Real))
function_table.add(Prototype("*", (Long, Real), Real))
function_table.add(Prototype("*", (Real, Integer), Real))
function_table.add(Prototype("*", (Integer, Real), Real))

function_table.add(Prototype("/", (Long, Long), Long))
function_table.add(Prototype("/", (Long, Integer), Long))
function_table.add(Prototype("/", (Integer, Long), Long))
function_table.add(Prototype("/", (Integer, Integer), Integer))
function_table.add(Prototype("/", (Real, Real), Real))
function_table.add(Prototype("/", (Real, Long), Real))
function_table.add(Prototype("/", (Long, Real), Real))
function_table.add(Prototype("/", (Real, Integer), Real))
function_table.add(Prototype("/", (Integer, Real), Real))

function_table.add(Prototype("-", (Timestamp, Timestamp), Real))

#
# Functions.
#
function_table.add(Prototype("covers", (Geometry, Geometry), Boolean))
function_table.add(Prototype("covers", (Timestamp, Timestamp, Timestamp, Timestamp), Boolean))
function_table.add(Prototype("distance", (Geometry, Geometry), Real))
function_table.add(Prototype("intersects", (Geometry, Geometry), Boolean))
function_table.add(Prototype("intersects", (Timestamp, Timestamp, Timestamp, Timestamp), Boolean))
function_table.add(Prototype("is_defined", (Long,), Boolean))
function_table.add(Prototype("is_defined", (Integer,), Boolean))
function_table.add(Prototype("is_defined", (Real,), Boolean))
function_table.add(Prototype("is_defined", (Boolean,), Boolean))
function_table.add(Prototype("is_defined", (Text,), Boolean))
function_table.add(Prototype("is_defined", (Namespace,), Boolean))
function_table.add(Prototype("is_defined", (Timestamp,), Boolean))
function_table.add(Prototype("is_defined", (UUID,), Boolean))
function_table.add(Prototype("is_defined", (Geometry,), Boolean))
function_table.add(Prototype("is_source_of", (UUID,), Boolean))
function_table.add(Prototype("is_source_of", (Boolean,), Boolean))
function_table.add(Prototype("is_derived_from", (UUID,), Boolean))
function_table.add(Prototype("is_derived_from", (Boolean,), Boolean))
function_table.add(Prototype("has_tag", (Text,), Boolean))
function_table.add(Prototype("now", (), Timestamp))


class TokenType(Enum):
    _items = ("TEXT", "UUID", "TIMESTAMP", "REAL", "INTEGER", "BOOLEAN", "NAME", "OPERATOR", "END")


class Token(object):
    def __init__(self, type_, value=None):
        self.type_ = type_
        self.value = value

    def __repr__(self):
        return "Token(type_ = TokenType.%s, value = %r)" % (TokenType.to_string(self.type_), self.value)


class TokenStream(object):
    _sub_patterns = \
        (
            r"""\"(?:[^\\"]|\\.)*\"""",                                      # Text literals
            r"""\d{4}-\d{2}-\d{2}(?:T\d{2}:\d{2}:\d{2}(?:\.\d{0,6})?)?""",   # Timestamp literals
            r"""[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}""",  # UUID literals
            r"""\d+(?:\.\d*(?:[eE][+-]?\d+)?|[eE][+-]?\d+)""",               # Real literals
            r"""0x[0-9a-fA-F]+|0o[0-7]\d+|0b[0-1]+|\d+""",                   # Integer literals
            r"""<=|>=|==|!=|~=|not in|[*<>@()\[\],.+-/]""",                  # Operators and delimiters
            r"""[a-zA-Z]\w*""",                                              # Names (incl. true, false, in)
        )

    _pattern = r"""(?:%s)""" % ("|".join(["(%s)" % sub_pattern for sub_pattern in _sub_patterns]))
    _re_token = re.compile(_pattern)

    _re_datemin = re.compile(r"0000-00-00(?:T00:00:00(?:\.0{0,6})?)?$")
    _re_datemax = re.compile(r"9999-99-99(?:T99:99:99(?:\.9{0,6})?)?$")

    def __init__(self, text):
        self.text = text
        self.at_end = not self.text
        self.token_start_position, self.token_end_position = 0, 0
        self.next()

    def next(self):
        if self.at_end:
            raise Error("char %d: unexpected end of input" % (self.token_start_position + 1))

        self.token = self._next_token()
        return self.token

    def test(self, types, values=None):
        return False if not self._test_token_types(types) else (values is None or self._test_token_values(values))

    def accept(self, types, values=None):
        if not self.test(types, values):
            return False

        self.next()
        return True

    def expect(self, types, values=None):
        if not self.test(types, values):
            if self.token.type_ == TokenType.END:
                raise Error("char %d: unexpected end of input" % (self.token_start_position + 1))
            else:
                if self.token.value is None:
                    token_str = TokenType.to_string(self.token.type_)
                else:
                    token_str = "\"%s\"" % self.token.value

                expected_str = self._types_to_string(types) if values is None else self._values_to_string(values)
                raise Error("char %d: expected %s, got %s" % (self.token_start_position + 1, expected_str, token_str))

        token = self.token
        self.next()
        return token

    def _types_to_string(self, types):
        try:
            strings = [TokenType.to_string(t) for t in types]
        except TypeError:
            return TokenType.to_string(types)

        return "%s%s" % ("" if len(strings) == 1 else "one of: ", ", ".join(strings))

    def _values_to_string(self, values):
        if isinstance(values, basestring):
            return "\"%s\"" % values

        try:
            strings = ["\"%s\"" % value for value in values]
        except TypeError:
            return "\"%s\"" % values

        return "%s%s" % ("" if len(strings) == 1 else "one of: ", ", ".join(strings))

    def _test_token_types(self, types):
        try:
            return self.token.type_ in types
        except TypeError:
            return self.token.type_ == types

    def _test_token_values(self, values):
        if isinstance(values, basestring):
            return self.token.value == values

        try:
            return self.token.value in values
        except TypeError:
            return self.token.value == values

    def _next_token(self):
        self.token_start_position = self._skip_white_space(self.token_end_position)

        if self.token_start_position == len(self.text):
            self.at_end = True
            return Token(TokenType.END)

        match_object = self._re_token.match(self.text, self.token_start_position)
        if match_object is None:
            raise Error("char %d: syntax error: \"%s\"" % (self.token_start_position + 1,
                        self.text[self.token_start_position:]))

        self.token_start_position, self.token_end_position = match_object.span()
        text, timestamp, uuid_, real, integer, operator, name = match_object.groups()

        if text is not None:
            return Token(TokenType.TEXT, string_unescape(text[1:-1]))

        if timestamp is not None:
            return Token(TokenType.TIMESTAMP, self._parse_timestamp(timestamp))

        if uuid_ is not None:
            return Token(TokenType.UUID, uuid.UUID(uuid_))

        if real is not None:
            return Token(TokenType.REAL, float(real))

        if integer is not None:
            if integer.startswith('0x'):
                base = 16
            elif integer.startswith('0o'):
                base = 8
            elif integer.startswith('0b'):
                base = 2
            else:
                base = 10
            return Token(TokenType.INTEGER, int(integer, base))

        if operator is not None:
            return Token(TokenType.OPERATOR, operator)

        if name is not None:
            if name in ["true", "false"]:
                return Token(TokenType.BOOLEAN, name == "true")
            elif name == "in":
                return Token(TokenType.OPERATOR, name)
            else:
                return Token(TokenType.NAME, name)

        raise Error("char %d: syntax error: \"%s\"" % (self.token_start_position + 1, match_object.group()))

    def _skip_white_space(self, start):
        while start < len(self.text) and self.text[start].isspace():
            start += 1
        return start

    def _parse_timestamp(self, timestamp):
        if self._re_datemin.match(timestamp) is not None:
            return datetime.datetime.min

        if self._re_datemax.match(timestamp) is not None:
            return datetime.datetime.max

        for format_string in ("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S.%f"):
            try:
                return datetime.datetime.strptime(timestamp, format_string)
            except ValueError:
                pass

        raise Error("char %d: invalid timestamp: \"%s\"" % (self.token_start_position + 1, timestamp))


class AbstractSyntaxTreeNode(object):
    pass


class Literal(AbstractSyntaxTreeNode):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.value)


class Name(AbstractSyntaxTreeNode):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.value)


class List(AbstractSyntaxTreeNode):
    def __init__(self, values):
        self.values = values

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.value)


class ParameterReference(AbstractSyntaxTreeNode):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.name)


class FunctionCall(AbstractSyntaxTreeNode):
    def __init__(self, name, *args):
        self.name = name
        self.arguments = list(args)

    def __str__(self):
        if not self.arguments:
            return "(%s %s)" % (type(self).__name__, self.name)
        return "(%s %s %s)" % (type(self).__name__, self.name, " ".join(map(str, self.arguments)))


def parse_sequence(stream, parse_item_function, start='(', end=')'):
    stream.expect(TokenType.OPERATOR, start)
    if stream.accept(TokenType.OPERATOR, end):
        return []

    sequence = [parse_item_function(stream)]
    while stream.accept(TokenType.OPERATOR, ","):
        sequence.append(parse_item_function(stream))
    stream.expect(TokenType.OPERATOR, end)
    return sequence


def parse_geometry_sequence(stream, parse_item_function):
    if stream.accept(TokenType.NAME, "EMPTY"):
        return []

    stream.expect(TokenType.OPERATOR, "(")
    sequence = [parse_item_function(stream)]
    while stream.accept(TokenType.OPERATOR, ","):
        sequence.append(parse_item_function(stream))
    stream.expect(TokenType.OPERATOR, ")")
    return sequence


def parse_signed_coordinate(stream):
    if stream.accept(TokenType.OPERATOR, "-"):
        token = stream.expect((TokenType.INTEGER, TokenType.REAL))
        return -float(token.value)

    stream.accept(TokenType.OPERATOR, "+")
    token = stream.expect((TokenType.INTEGER, TokenType.REAL))
    return float(token.value)


def parse_point_raw(stream):
    return geometry.Point(parse_signed_coordinate(stream),
                          parse_signed_coordinate(stream))


def parse_point(stream):
    stream.expect(TokenType.OPERATOR, "(")
    point = parse_point_raw(stream)
    stream.expect(TokenType.OPERATOR, ")")
    return point


def parse_line_string(stream):
    return geometry.LineString(parse_geometry_sequence(stream, parse_point_raw))


def parse_linear_ring(stream):
    points = parse_geometry_sequence(stream, parse_point_raw)
    if len(points) == 0:
        return geometry.LinearRing()

    if len(points) < 4:
        raise Error("char %d: linear ring should be empty or should contain >= 4 points" % stream.token_start_position)

    if points[-1] != points[0]:
        raise Error("char %d: linear ring should be closed" % stream.token_start_position)

    return geometry.LinearRing(points)


def parse_polygon(stream):
    return geometry.Polygon(parse_geometry_sequence(stream, parse_linear_ring))


def parse_multi_point(stream):
    return geometry.MultiPoint(parse_geometry_sequence(stream, parse_point))


def parse_multi_line_string(stream):
    return geometry.MultiLineString(parse_geometry_sequence(stream, parse_line_string))


def parse_multi_polygon(stream):
    return geometry.MultiPolygon(parse_geometry_sequence(stream, parse_polygon))


def parse_atom(stream):
    # Sub-expression.
    if stream.accept(TokenType.OPERATOR, "("):
        sub_expression = parse_expression(stream)
        stream.expect(TokenType.OPERATOR, ")")
        return sub_expression

    # Parameter reference.
    if stream.accept(TokenType.OPERATOR, "@"):
        name_token = stream.expect(TokenType.NAME)
        return ParameterReference(name_token.value)

    # Geometry literal, function call, or name.
    if stream.test(TokenType.NAME):
        name_token = stream.expect(TokenType.NAME)

        # Geometry literals.
        if name_token.value == "POINT":
            return Literal(parse_point(stream))
        elif name_token.value == "LINESTRING":
            return Literal(parse_line_string(stream))
        elif name_token.value == "POLYGON":
            return Literal(parse_polygon(stream))
        elif name_token.value == "MULTIPOINT":
            return Literal(parse_multi_point(stream))
        elif name_token.value == "MULTILINESTRING":
            return Literal(parse_multi_line_string(stream))
        elif name_token.value == "MULTIPOLYGON":
            return Literal(parse_multi_polygon(stream))

        # Function call.
        if stream.test(TokenType.OPERATOR, "("):
            return FunctionCall(name_token.value, *parse_sequence(stream, parse_expression))

        # Name (possibly qualified).
        parts = [name_token.value]
        while stream.accept(TokenType.OPERATOR, "."):
            name_token = stream.expect(TokenType.NAME)
            parts.append(name_token.value)
        return Name(".".join(parts))

    if stream.test(TokenType.OPERATOR, "["):
        return List(parse_sequence(stream, parse_expression, "[", "]"))

    # Literal.
    token = stream.expect((TokenType.TEXT, TokenType.TIMESTAMP, TokenType.UUID, TokenType.REAL, TokenType.INTEGER,
                           TokenType.BOOLEAN))
    return Literal(token.value)


def parse_term(stream):
    if stream.test(TokenType.OPERATOR, ("+", "-")):
        operator_token = stream.expect(TokenType.OPERATOR, ("+", "-"))
        return FunctionCall(operator_token.value, parse_term(stream))
    return parse_atom(stream)


def parse_arithmetic_expression(stream):
    lhs = parse_term(stream)
    if stream.test(TokenType.OPERATOR, ("+", "-", "*", "/")):
        operator_token = stream.expect(TokenType.OPERATOR, ("+", "-", "*", "/"))
        return FunctionCall(operator_token.value, lhs, parse_arithmetic_expression(stream))
    return lhs


def parse_comparison(stream):
    lhs = parse_arithmetic_expression(stream)
    if stream.test(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=", "in", "not in")):
        operator_token = stream.expect(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=", "in", "not in"))
        return FunctionCall(operator_token.value, lhs, parse_comparison(stream))
    return lhs


def parse_not_expression(stream):
    if stream.accept(TokenType.NAME, "not"):
        return FunctionCall("not", parse_not_expression(stream))
    return parse_comparison(stream)


def parse_and_expression(stream):
    lhs = parse_not_expression(stream)
    if stream.accept(TokenType.NAME, "and"):
        return FunctionCall("and", lhs, parse_and_expression(stream))
    return lhs


def parse_or_expression(stream):
    lhs = parse_and_expression(stream)
    if stream.accept(TokenType.NAME, "or"):
        return FunctionCall("or", lhs, parse_or_expression(stream))
    return lhs


def parse_expression(stream):
    return parse_or_expression(stream)


def _literal_type(literal):
    if isinstance(literal, list):  # TODO use Sequence.validate..?
        return Sequence

    for type in (Text, Timestamp, UUID, Boolean, Integer, Long, Real, Geometry):
        try:
            type.validate(literal)
        except ValueError:
            pass
        else:
            return type

    raise Error("unable to determine type of literal value: %r" % (literal,))


class SemanticAnalysis(Visitor):
    def __init__(self, namespace_schemas, parameters, having=False):
        super(SemanticAnalysis, self).__init__()
        self._namespace_schemas = namespace_schemas
        self._parameters = parameters
        self._having = having

    def visit_Literal(self, visitable):
        visitable.type = _literal_type(visitable.value)

    def visit_Name(self, visitable):
        if self._having:
            item = Identifier(visitable.value, self._namespace_schemas)
            visitable.type = item.muninn_type
            visitable.value = item
            return

        split_name = visitable.value.split(".")

        # namespace/implicit core property
        if len(split_name) == 1:
            if split_name[0] in self._namespace_schemas:
                namespace = split_name[0]
                name = None
            else:
                namespace, name = "core", split_name[0]

        # namespace.property
        elif len(split_name) == 2:
            namespace, name = split_name

        else:
            raise Error("invalid property name: \"%s\"" % visitable.value)

        # check that namespace exists
        try:
            schema = self._namespace_schemas[namespace]
        except KeyError:
            raise Error("undefined namespace: \"%s\"" % namespace)

        # namespace
        if name is None:
            visitable.value = split_name[0]
            visitable.type = Namespace

        # namespace.property
        else:
            try:
                type_ = schema[name]
            except KeyError:
                if len(split_name) == 2:
                    raise Error("undefined property: \"%s\"" % visitable.value)
                else:
                    raise Error("undefined name: \"%s\"" % name)

            visitable.value = "%s.%s" % (namespace, name)
            visitable.type = type_

    def visit_List(self, visitable):  # TODO check same literal type
        values = []
        for value in visitable.values:
            if not isinstance(value, Literal):
                raise Error("list contains non-literal")
            values.append(value.value)
        visitable.value = values
        visitable.type = Sequence

    def visit_ParameterReference(self, visitable):
        try:
            value = self._parameters[visitable.name]
        except KeyError:
            raise Error("no value for parameter: \"%s\"" % visitable.name)

        visitable.value = value
        visitable.type = _literal_type(value)

    def visit_FunctionCall(self, visitable):
        # Resolve the type of the function arguments.
        for argument in visitable.arguments:
            self.visit(argument)

        prototype = Prototype(visitable.name, [argument.type for argument in visitable.arguments])

        try:
            prototypes = function_table.resolve(prototype)
        except KeyError:
            prototypes = []

        if not prototypes:
            raise Error("undefined function: \"%s\"" % prototype)

        if len(prototypes) > 1:
            raise InternalError("cannot uniquely resolve function: \"%s\"" % prototype)

        prototype = prototypes[0]
        visitable.prototype = prototype
        visitable.type = prototype.return_type

    def visit_AbstractSyntaxTreeNode(self, visitable):
        if not hasattr(visitable, "type"):
            raise InternalError("encountered abstract syntax tree node without type attribute: %s" %
                                type(visitable).__name__)

    def default(self, visitable):
        raise InternalError("unsupported abstract syntax tree node type: %s" % type(visitable).__name__)


def parse(text):
    stream = TokenStream(text)
    abstract_syntax_tree = parse_expression(stream)
    if not stream.test(TokenType.END):
        raise Error("char %d: extra characters after expression: \"%s\"" % (stream.token_start_position + 1,
                                                                            text[stream.token_start_position:]))
    return abstract_syntax_tree


def analyze(abstract_syntax_tree, namespace_schemas={}, parameters={}, having=False):
    annotated_syntax_tree = copy.deepcopy(abstract_syntax_tree)
    SemanticAnalysis(namespace_schemas, parameters, having=having).visit(annotated_syntax_tree)
    return annotated_syntax_tree


def parse_and_analyze(text, namespace_schemas={}, parameters={}, having=False):
    return analyze(parse(text), namespace_schemas, parameters, having)


def string_unescape(text):
    '''
    Unescape special characters in a string.
    Python2 and 3 compatible, uses the native string type.
    In python2, the same effect can also be achieved with `string.decode("string-escape")`
    '''
    text = str(text)  # make sure we are using the native string type
    regex = re.compile('\\\\(\\\\|[\'"abfnrtv])')
    translator = {
        '\\': '\\',
        "'": "'",
        '"': '"',
        'a': '\a',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        'v': '\v',
    }

    def _replace(m):
        c = m.group(1)
        return translator[c]

    result = regex.sub(_replace, text)
    return result


class Identifier(object):
    def __init__(self, canonical_identifier, namespace_schemas):
        self.canonical = canonical_identifier
        self.subscript = None

        if canonical_identifier == 'tag':
            # the rules to get the namespace database table name also apply to 'tag'
            self.namespace = canonical_identifier
            self.identifier = canonical_identifier
            self.muninn_type = Text

        elif canonical_identifier == 'count':
            self.namespace = None
            self.identifier = canonical_identifier
            self.muninn_type = Long

        elif not re.match(r'[\w]+\.[\w.]+', canonical_identifier):
            raise Error("cannot resolve identifier: %r" % (canonical_identifier,))

        else:
            segments = canonical_identifier.split('.')

            if len(segments) == 1:
                self.namespace = 'core'
                self.identifier = segments[0]

            elif len(segments) == 2:
                if segments[0] in namespace_schemas:
                    self.namespace, self.identifier = segments
                else:
                    self.namespace = 'core'
                    self.identifier, self.subscript = segments

            elif len(segments) == 3:
                self.namespace, self.identifier, self.subscript = segments

            else:
                raise Error("cannot resolve identifier: %r" % (canonical_identifier,))

            # check if namespace is valid
            if self.namespace not in namespace_schemas:
                raise Error("undefined namespace: \"%s\"" % self.namespace)

            # check if property name is valid
            if self.identifier not in namespace_schemas[self.namespace]:
                if self.property_name != 'core.validity_duration':
                    raise Error("no property: %r defined within namespace: %r" % (self.identifier, self.namespace))

            # note: not checking if subscript is valid; the list of possible subscripts varies depending on context
            if self.property_name == 'core.validity_duration':
                self.muninn_type = None
            else:
                self.muninn_type = namespace_schemas[self.namespace][self.identifier]

    @property
    def property_name(self):
        return '%s.%s' % (self.namespace, self.identifier)

    @property
    def resolve(self):
        if self.canonical in ('count', 'tag'):
            return self.canonical
        elif self.subscript is None:
            return '%s.%s' % (self.namespace, self.identifier)
        else:
            return '%s.%s.%s' % (self.namespace, self.identifier, self.subscript)



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import muninn

from muninn.tools.utils import Processor, create_parser, parse_args_and_run


class UntagProcessor(Processor):
    def __init__(self, args):
        super(UntagProcessor, self).__init__(args)
        self.args = args

    def perform_operation(self, archive, product):
        tags = None if self.args.all else self.args.tag
        archive.untag(product.core.uuid, tags)
        return 1


def untag(args):
    processor = UntagProcessor(args)
    with muninn.open(args.archive) as archive:
        products = archive.search(where=args.expression, property_names=['uuid'])
        return processor.process(archive, args, products)


def main():
    parser = create_parser(description="Remove one or more tags from products contained in a muninn archive",
                           parallel=True)
    parser.add_argument("-a", "--all", action="store_true",
                        help="ignore tags supplied on the command line; instead remove all tags")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", help="expression used to search for products to untag")
    parser.add_argument("tag", metavar="TAG", nargs="*", help="tags to remove")
    return parse_args_and_run(parser, untag)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import muninn

from muninn.tools.utils import create_parser, parse_args_and_run


def run(args):
    archive_names = args.archive or muninn.list_archives()

    print("ARCHIVES")
    for index, archive_name in enumerate(sorted(archive_names)):
        if index > 0 and not args.name_only:
            print()
        print("  " + archive_name)
        try:
            with muninn.open(archive_name) as archive:
                if args.name_only:
                    continue
                print("    NAMESPACES")
                for namespace in sorted(archive.namespaces()):
                    print("      %s" % namespace)
                    namespace_schema = archive.namespace_schema(namespace)
                    for name in sorted(namespace_schema):
                        field = namespace_schema[name]
                        field_name = field.name()
                        if field.__module__ != 'muninn.schema':
                            field_name = '%s.%s' % (field.__module__, field.name())
                        optional = namespace_schema.is_optional(name)
                        print("        %s: %s%s" % (name, field_name, " (optional)" if optional else "", ))

                print("\n    PRODUCT TYPES")
                for product_type in sorted(archive.product_types()):
                    print("      %s" % product_type)

                if archive.remote_backends():
                    print("\n    REMOTE BACKENDS")
                    for remote_backend in sorted(archive.remote_backends()):
                        print("      %s" % remote_backend)

                if archive.synchronizers():
                    print("\n    SYNCHRONIZERS")
                    for synchronizer in sorted(archive.synchronizers()):
                        print("      %s" % synchronizer)
        except Exception:
            print('    (could not open archive)')

    return 0


def main():
    parser = create_parser(description="Display generic archive information")
    parser.add_argument("-n", "--name-only", action="store_true",
                        help="only show archive name")
    parser.add_argument("archive", nargs='*', metavar="ARCHIVE",
                        help="archive identifier (default: search and show all archives)")
    return parse_args_and_run(parser, run)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import collections

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

import inspect

from muninn.exceptions import Error
from muninn.function import Prototype
from muninn.language import parse_and_analyze, Literal, Name, Identifier
from muninn.schema import Boolean, Integer, Long, Real, Sequence, Text, Timestamp, UUID
from muninn.visitor import Visitor
from muninn.core import Basename, ArchivePath


AGGREGATE_FUNCTIONS = collections.OrderedDict([
    (Long, ['min', 'max', 'sum', 'avg']),
    (Integer, ['min', 'max', 'sum', 'avg']),
    (Real, ['min', 'max', 'sum', 'avg']),
    # (Boolean, []),
    (Text, ['min', 'max']),
    (Timestamp, ['min', 'max']),
    # (UUID, []),
    # (Geometry, []),
    (None, ['min', 'max', 'sum', 'avg']),  # special case: validity_duration
])

GROUP_BY_FUNCTIONS = collections.OrderedDict([
    (Long, [None, ]),
    (Integer, [None, ]),
    # (Real, []),
    (Boolean, [None, ]),
    (Text, [None, 'length']),
    (Basename, [None, 'length']),
    (ArchivePath, [None, 'length']),
    (Timestamp, ['year', 'month', 'yearmonth', 'date', 'day', 'hour', 'minute', 'second', 'time']),
    # (UUID, []),
    # (Geometry, []),
])


class TypeMap(MutableMapping):
    def __init__(self):
        self._types = {}

    def __getitem__(self, key):
        for type in inspect.getmro(key):
            try:
                return self._types[type]
            except KeyError:
                pass

        raise KeyError(key.__name__)

    def __setitem__(self, key, value):
        self._types[key] = value

    def __delitem__(self, key):
        del self._types[key]

    def __iter__(self):
        return iter(self._types)

    def __len__(self):
        return len(self._types)


def as_is(sql):
    return lambda: sql


def unary_operator_rewriter(operator):
    return lambda arg0: "%s (%s)" % (operator, arg0)


def binary_operator_rewriter(operator):
    return lambda arg0, arg1: "(%s) %s (%s)" % (arg0, operator, arg1)


def unary_function_rewriter(name):
    return lambda arg0: "%s(%s)" % (name, arg0)


def membership_operator_rewriter(operator):
    return lambda arg0, arg1: "(%s) %s %s" % (arg0, operator, arg1)


def binary_function_rewriter(name):
    return lambda arg0, arg1: "%s(%s, %s)" % (name, arg0, arg1)


def default_rewriter_table():
    #
    # Table of all supported operators and functions that can be rewritten in standard SQL and without backend specific
    # knowledge. NB. This is a subset of the set of all supported operators and functions. Database specific backends
    # should add custom implementations of the operators and functions not implemented here.
    #
    rewriter_table = {}

    #
    # Logical operators
    #
    rewriter_table[Prototype("not", (Boolean,), Boolean)] = unary_operator_rewriter("NOT")
    rewriter_table[Prototype("and", (Boolean, Boolean), Boolean)] = binary_operator_rewriter("AND")
    rewriter_table[Prototype("or", (Boolean, Boolean), Boolean)] = binary_operator_rewriter("OR")

    #
    # Membership operators
    #
    in_rewriter = membership_operator_rewriter("in")
    rewriter_table[Prototype("in", (Integer, Sequence), Boolean)] = in_rewriter
    rewriter_table[Prototype("in", (Long, Sequence), Boolean)] = in_rewriter
    rewriter_table[Prototype("in", (Real, Sequence), Boolean)] = in_rewriter
    rewriter_table[Prototype("in", (Text, Sequence), Boolean)] = in_rewriter
    not_in_rewriter = membership_operator_rewriter("not in")
    rewriter_table[Prototype("not in", (Integer, Sequence), Boolean)] = not_in_rewriter
    rewriter_table[Prototype("not in", (Long, Sequence), Boolean)] = not_in_rewriter
    rewriter_table[Prototype("not in", (Real, Sequence), Boolean)] = not_in_rewriter
    rewriter_table[Prototype("not in", (Text, Sequence), Boolean)] = not_in_rewriter

    #
    # Comparison operators
    #
    eq_rewriter = binary_operator_rewriter("=")
    rewriter_table[Prototype("==", (Long, Long), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Long, Integer), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Integer, Long), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Integer, Integer), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Real, Real), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Real, Long), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Long, Real), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Real, Integer), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Integer, Real), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Boolean, Boolean), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Text, Text), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (Timestamp, Timestamp), Boolean)] = eq_rewriter
    rewriter_table[Prototype("==", (UUID, UUID), Boolean)] = eq_rewriter

    ne_rewriter = binary_operator_rewriter("!=")
    rewriter_table[Prototype("!=", (Long, Long), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Long, Integer), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Integer, Long), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Integer, Integer), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Real, Real), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Real, Long), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Long, Real), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Real, Integer), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Integer, Real), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Boolean, Boolean), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Text, Text), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (Timestamp, Timestamp), Boolean)] = ne_rewriter
    rewriter_table[Prototype("!=", (UUID, UUID), Boolean)] = ne_rewriter

    lt_rewriter = binary_operator_rewriter("<")
    rewriter_table[Prototype("<", (Long, Long), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Long, Integer), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Integer, Long), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Integer, Integer), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Real, Real), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Real, Long), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Long, Real), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Real, Integer), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Integer, Real), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Text, Text), Boolean)] = lt_rewriter
    rewriter_table[Prototype("<", (Timestamp, Timestamp), Boolean)] = lt_rewriter

    gt_rewriter = binary_operator_rewriter(">")
    rewriter_table[Prototype(">", (Long, Long), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Long, Integer), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Integer, Long), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Integer, Integer), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Real, Real), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Real, Long), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Long, Real), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Real, Integer), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Integer, Real), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Text, Text), Boolean)] = gt_rewriter
    rewriter_table[Prototype(">", (Timestamp, Timestamp), Boolean)] = gt_rewriter

    le_rewriter = binary_operator_rewriter("<=")
    rewriter_table[Prototype("<=", (Long, Long), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Long, Integer), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Integer, Long), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Integer, Integer), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Real, Real), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Real, Long), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Long, Real), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Real, Integer), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Integer, Real), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Text, Text), Boolean)] = le_rewriter
    rewriter_table[Prototype("<=", (Timestamp, Timestamp), Boolean)] = le_rewriter

    ge_rewriter = binary_operator_rewriter(">=")
    rewriter_table[Prototype(">=", (Long, Long), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Long, Integer), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Integer, Long), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Integer, Integer), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Real, Real), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Real, Long), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Long, Real), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Real, Integer), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Integer, Real), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Text, Text), Boolean)] = ge_rewriter
    rewriter_table[Prototype(">=", (Timestamp, Timestamp), Boolean)] = ge_rewriter

    rewriter_table[Prototype("~=", (Text, Text), Boolean)] = binary_operator_rewriter("LIKE")

    #
    # Arithmetic operators.
    #
    plus_rewriter = unary_operator_rewriter("+")
    rewriter_table[Prototype("+", (Long,), Long)] = plus_rewriter
    rewriter_table[Prototype("+", (Integer,), Integer)] = plus_rewriter
    rewriter_table[Prototype("+", (Real,), Real)] = plus_rewriter

    minus_rewriter = unary_operator_rewriter("-")
    rewriter_table[Prototype("-", (Long,), Long)] = minus_rewriter
    rewriter_table[Prototype("-", (Integer,), Integer)] = minus_rewriter
    rewriter_table[Prototype("-", (Real,), Real)] = minus_rewriter

    add_rewriter = binary_operator_rewriter("+")
    rewriter_table[Prototype("+", (Long, Long), Long)] = add_rewriter
    rewriter_table[Prototype("+", (Long, Integer), Long)] = add_rewriter
    rewriter_table[Prototype("+", (Integer, Long), Long)] = add_rewriter
    rewriter_table[Prototype("+", (Integer, Integer), Integer)] = add_rewriter
    rewriter_table[Prototype("+", (Real, Real), Real)] = add_rewriter
    rewriter_table[Prototype("+", (Real, Long), Real)] = add_rewriter
    rewriter_table[Prototype("+", (Long, Real), Real)] = add_rewriter
    rewriter_table[Prototype("+", (Real, Integer), Real)] = add_rewriter
    rewriter_table[Prototype("+", (Integer, Real), Real)] = add_rewriter

    subtract_rewriter = binary_operator_rewriter("-")
    rewriter_table[Prototype("-", (Long, Long), Long)] = subtract_rewriter
    rewriter_table[Prototype("-", (Long, Integer), Long)] = subtract_rewriter
    rewriter_table[Prototype("-", (Integer, Long), Long)] = subtract_rewriter
    rewriter_table[Prototype("-", (Integer, Integer), Integer)] = subtract_rewriter
    rewriter_table[Prototype("-", (Real, Real), Real)] = subtract_rewriter
    rewriter_table[Prototype("-", (Real, Long), Real)] = subtract_rewriter
    rewriter_table[Prototype("-", (Long, Real), Real)] = subtract_rewriter
    rewriter_table[Prototype("-", (Real, Integer), Real)] = subtract_rewriter
    rewriter_table[Prototype("-", (Integer, Real), Real)] = subtract_rewriter

    multiply_rewriter = binary_operator_rewriter("*")
    rewriter_table[Prototype("*", (Long, Long), Long)] = multiply_rewriter
    rewriter_table[Prototype("*", (Long, Integer), Long)] = multiply_rewriter
    rewriter_table[Prototype("*", (Integer, Long), Long)] = multiply_rewriter
    rewriter_table[Prototype("*", (Integer, Integer), Integer)] = multiply_rewriter
    rewriter_table[Prototype("*", (Real, Real), Real)] = multiply_rewriter
    rewriter_table[Prototype("*", (Real, Long), Real)] = multiply_rewriter
    rewriter_table[Prototype("*", (Long, Real), Real)] = multiply_rewriter
    rewriter_table[Prototype("*", (Real, Integer), Real)] = multiply_rewriter
    rewriter_table[Prototype("*", (Integer, Real), Real)] = multiply_rewriter

    divide_rewriter = binary_operator_rewriter("/")
    rewriter_table[Prototype("/", (Long, Long), Long)] = divide_rewriter
    rewriter_table[Prototype("/", (Long, Integer), Long)] = divide_rewriter
    rewriter_table[Prototype("/", (Integer, Long), Long)] = divide_rewriter
    rewriter_table[Prototype("/", (Integer, Integer), Integer)] = divide_rewriter
    rewriter_table[Prototype("/", (Real, Real), Real)] = divide_rewriter
    rewriter_table[Prototype("/", (Real, Long), Real)] = divide_rewriter
    rewriter_table[Prototype("/", (Long, Real), Real)] = divide_rewriter
    rewriter_table[Prototype("/", (Real, Integer), Real)] = divide_rewriter
    rewriter_table[Prototype("/", (Integer, Real), Real)] = divide_rewriter

    #
    # Functions.
    #
    rewriter_table[Prototype("covers", (Timestamp, Timestamp, Timestamp, Timestamp), Boolean)] = \
        lambda left0, right0, left1, right1: "(%s) >= (%s) AND (%s) >= (%s) AND (%s) >= (%s) AND (%s) <= (%s)" % \
        (right0, left0, right1, left1, left1, left0, right1, right0)

    rewriter_table[Prototype("intersects", (Timestamp, Timestamp, Timestamp, Timestamp), Boolean)] = \
        lambda left0, right0, left1, right1: "(%s) >= (%s) AND (%s) >= (%s) AND (%s) >= (%s) AND (%s) <= (%s)" % \
        (right0, left0, right1, left1, right0, left1, left0, right1)

    return rewriter_table


class _WhereExpressionVisitor(Visitor):
    def __init__(self, rewriter_table, column_name_func, named_placeholder_func, root_visitor=None):
        super(_WhereExpressionVisitor, self).__init__()
        self._rewriter_table = rewriter_table
        self._column_name = column_name_func
        self._named_placeholder = named_placeholder_func

        self._root_visitor = root_visitor or self

    def do_visit(self, visitable):
        where, parameters, namespaces = self.visit(visitable)
        if isinstance(visitable, Literal) and visitable.type is UUID:
            where = ("(uuid = %s)" % where)
        return where, parameters, namespaces

    def visit(self, visitable):
        self._count, self._parameters, self._namespaces = 0, {}, set()
        return super(_WhereExpressionVisitor, self).visit(visitable), self._parameters, self._namespaces

    def visit_Literal(self, visitable):
        parameter_name = str(self._root_visitor._count)
        self._root_visitor._parameters[parameter_name] = visitable.value
        self._root_visitor._count += 1
        return self._named_placeholder(parameter_name, arg=visitable.value)

    def visit_Name(self, visitable):
        if isinstance(visitable.value, Identifier):
            item = visitable.value
            if item.canonical == 'count':
                return 'COUNT(*)'
            else:
                return '%s(%s)' % (item.subscript.upper(), self._column_name(item.namespace, item.identifier))

        namespace_name = visitable.value.split('.')
        if len(namespace_name) == 1:
            namespace = namespace_name[0]
            name = namespace
        else:
            namespace, name = namespace_name
            name = self._column_name(namespace, name)
        self._namespaces.add(namespace)
        return name

    def visit_List(self, visitable):  # TODO parameters?
        return '(' + ','.join(repr(v) for v in visitable.value) + ')'

    def visit_ParameterReference(self, visitable):
        if visitable.type is Sequence:
            result = []
            for value in visitable.value:
                parameter_name = str(self._root_visitor._count)
                self._root_visitor._parameters[parameter_name] = value
                self._root_visitor._count += 1
                result.append(self._named_placeholder(parameter_name, arg=value))
            return '(' + ','.join(result) + ')'
        else:
            parameter_name = str(self._root_visitor._count)
            self._root_visitor._parameters[parameter_name] = visitable.value
            self._root_visitor._count += 1
            return self._named_placeholder(parameter_name, arg=visitable.value)

    def visit_FunctionCall(self, visitable):
        try:
            rewriter_func = self._rewriter_table[visitable.prototype]
        except KeyError:
            raise Error("function not supported by backend: %s" % visitable.prototype)

        # sub-query
        prototype = visitable.prototype
        if (prototype.name in ('is_source_of', 'is_derived_from') and prototype.argument_types and
                prototype.argument_types[0].name() == 'boolean'):

            visitor = _WhereExpressionVisitor(self._rewriter_table, self._column_name, self._named_placeholder,
                                              self._root_visitor)
            where_expr, where_parameters, where_namespaces = visitor.visit(visitable.arguments[0])

            if 'core' in where_namespaces:
                where_namespaces.remove('core')

            return rewriter_func(where_expr, where_namespaces)

        # non-sub-query
        else:
            arguments = []
            for type_, argument in zip(visitable.prototype.argument_types, visitable.arguments):
                where = super(_WhereExpressionVisitor, self).visit(argument)
                if isinstance(argument, Literal) and argument.type is UUID and type_ is Boolean:
                    where = ("(uuid = %s)" % where)
                arguments.append(where)
            sql_expr = rewriter_func(*arguments)

            # SQL-improved NULL checking (e.g., field != "blah" also matches NULL)
            if visitable.name in ('==', '!=', '~='):
                if (isinstance(visitable.arguments[0], Name) and isinstance(visitable.arguments[1], Literal)):
                    name = arguments[0]
                elif (isinstance(visitable.arguments[0], Literal) and isinstance(visitable.arguments[1], Name)):
                    name = arguments[1]
                else:
                    return sql_expr

                if visitable.name == '!=':
                    return '(%s OR %s IS NULL)' % (sql_expr, name)
                else:
                    return '(%s AND %s IS NOT NULL)' % (sql_expr, name)

            return sql_expr

    def default(self, visitable):
        raise Error("unsupported abstract syntax tree node type: %r" % type(visitable).__name__)


class SQLBuilder(object):
    def __init__(self, namespace_schemas, type_map, rewriter_table, table_name_func, _named_placeholder_func,
                 _placeholder_func, rewriter_property_func):
        self._namespace_schemas = namespace_schemas
        self._type_map = type_map
        self._rewriter_table = rewriter_table
        self._table_name = table_name_func
        self._named_placeholder = _named_placeholder_func
        self._placeholder = _placeholder_func
        self._rewriter_property = rewriter_property_func

    def build_create_table_query(self, namespace):
        column_sql = []

        schema = self._namespace_schema(namespace)
        for identifier in schema:
            sql = identifier + " " + self._type(schema[identifier])
            if not schema.is_optional(identifier):
                sql = sql + " " + "NOT NULL"
            column_sql.append(sql)

        return "CREATE TABLE %s (%s);" % (self._table_name(namespace), ", ".join(column_sql))

    def build_count_query(self, where="", parameters={}):
        join_set = set()

        # Parse the where clause.
        where_clause, where_parameters = "", {}
        if where:
            ast = parse_and_analyze(where, self._namespace_schemas, parameters)
            visitor = _WhereExpressionVisitor(self._rewriter_table, self._column_name, self._named_placeholder)
            where_expr, where_parameters, where_namespaces = visitor.do_visit(ast)
            if where_expr:
                join_set.update(where_namespaces)
                where_clause = "WHERE %s" % where_expr

        # Generate the FROM clause.
        from_clause = "FROM %s" % self._table_name("core")

        join_set.discard("core")
        for namespace in join_set:
            from_clause = "%s LEFT JOIN %s USING (uuid)" % (from_clause, self._table_name(namespace))

        # Generate the complete query.
        query = "SELECT COUNT(*) AS count %s" % from_clause
        if where_clause:
            query = "%s %s" % (query, where_clause)

        return query, where_parameters

    def build_summary_query(self, where='', parameters=None, aggregates=None, group_by=None, group_by_tag=False,
                            having=None, order_by=None):
        aggregates = aggregates or []
        group_by = group_by or []
        order_by = order_by or []
        if group_by_tag:
            group_by = group_by + ['tag']

        join_set = set()
        result_fields = []

        for field in group_by + ['count'] + aggregates:
            ident = Identifier(field, self._namespace_schemas)
            result_fields.append(ident.resolve)
            join_set.add(ident.namespace)

        # Parse the WHERE clause.
        where_clause, where_parameters = '', {}
        if where:
            ast = parse_and_analyze(where, self._namespace_schemas, parameters)
            visitor = _WhereExpressionVisitor(self._rewriter_table, self._column_name, self._named_placeholder)
            where_expr, where_parameters, where_namespaces = visitor.do_visit(ast)
            if where_expr:
                join_set.update(where_namespaces)
                where_clause = 'WHERE %s' % where_expr

        # Generate the GROUP BY clause.
        group_by_clause = ''
        group_by_list = [str(i) for i in range(1, len(group_by) + 1)]
        if group_by_list:
            group_by_clause = 'GROUP BY %s' % ', '.join(group_by_list)

        # Generate the HAVING clause
        having_clause = ''
        if having is not None:
            ast = parse_and_analyze(having, self._namespace_schemas, parameters, having=True)
            visitor = _WhereExpressionVisitor(self._rewriter_table, self._column_name, self._named_placeholder,
                                              visitor if where else None)
            having_expr, having_parameters, _ = visitor.do_visit(ast)
            where_parameters.update(having_parameters)
            having_clause = 'HAVING %s' % having_expr

        # Parse the ORDER BY clause.
        order_by_clause = ''
        order_by_list = []
        for item in order_by:
            direction = 'DESC' if item.startswith('-') else 'ASC'
            name = item[1:] if item.startswith('+') or item.startswith('-') else item
            name = Identifier(name, self._namespace_schemas).resolve
            if name not in result_fields:
                raise Error("cannot order result by %r; field is not present in result" % name)
            order_by_list.append('"%s" %s' % (name, direction))
        order_by_list += [str(i) for i in range(1, len(group_by) + 1)]
        if order_by_list:
            order_by_clause = 'ORDER BY %s' % ', '.join(order_by_list)

        # Generate the SELECT clause.
        select_list = []
        # group by fields
        for item in group_by:
            item = Identifier(item, self._namespace_schemas)
            column_name = self._column_name(item.namespace, item.identifier)
            group_by_functions = GROUP_BY_FUNCTIONS.get(item.muninn_type)
            if not group_by_functions:  # item.muninn_type not in (Text, Boolean, Long, Integer):
                if item.muninn_type:
                    raise Error("property %r of type %r cannot be part of the group_by field specification" %
                                (item.property_name, item.muninn_type.name()))
                else:
                    raise Error("property %r cannot be part of the group_by field specification" %
                                (item.property_name, ))
            if item.subscript not in group_by_functions:
                if item.subscript:
                    allowed_message = "; it can be one of %r" % group_by_functions if group_by_functions != [None] \
                        else ""
                    raise Error(("group field specification subscript %r of %r is not allowed" + allowed_message) %
                                (item.subscript, item.canonical))
                else:
                    raise Error("property %r of type %r must specify a subscript (one of %r) to be part of the "
                                "group_by field specification" % (item.property_name, item.muninn_type.name(),
                                                                  group_by_functions))
            if item.subscript:
                column_name = self._rewriter_property(column_name, item.subscript)
            select_list.append('%s AS "%s"' % (column_name, item.resolve))

        # aggregated fields
        select_list.append('COUNT(*) AS count')  # always aggregate row count
        for item in aggregates:
            item = Identifier(item, self._namespace_schemas)
            join_set.add(item.namespace)

            if not AGGREGATE_FUNCTIONS.get(item.muninn_type):
                raise Error("property %r of type %r cannot be part of the summary field specification" %
                            (item.property_name, item.muninn_type.name()))
            elif item.subscript not in AGGREGATE_FUNCTIONS[item.muninn_type]:
                if item.subscript:
                    raise Error("summary field specification subscript %r of %r should be one of %r" %
                                (item.subscript, item.canonical, AGGREGATE_FUNCTIONS[item.muninn_type]))
                else:
                    raise Error("summary field specification %r must specify a subscript (one of %r)" %
                                (item.canonical, AGGREGATE_FUNCTIONS[item.muninn_type]))
            if item.property_name == 'core.validity_duration':
                start_column = self._column_name(item.namespace, 'validity_start')
                stop_column = self._column_name(item.namespace, 'validity_stop')
                column_name = self._rewriter_table[Prototype('-', (Timestamp, Timestamp), Real)](stop_column,
                                                                                                 start_column)
            else:
                column_name = self._column_name(item.namespace, item.identifier)
            select_list.append('%s(%s) AS "%s"' % (item.subscript.upper(), column_name, item.canonical))
        select_clause = 'SELECT %s' % ', '.join(select_list)

        # Generate the FROM clause.
        from_clause = 'FROM %s' % self._table_name('core')

        for namespace in join_set:
            if namespace not in ('core', None):
                from_clause = '%s LEFT JOIN %s USING (uuid)' % (from_clause, self._table_name(namespace))

        # Generate the complete query.
        query = '%s\n%s' % (select_clause, from_clause)
        if where_clause:
            query = '%s\n%s' % (query, where_clause)
        if group_by_clause:
            query = '%s\n%s' % (query, group_by_clause)
        if having_clause:
            query = '%s\n%s' % (query, having_clause)
        if order_by_clause:
            query = '%s\n%s' % (query, order_by_clause)

        return query, where_parameters, result_fields

    def build_search_query(self, where="", order_by=[], limit=None, parameters={}, namespaces=[], property_names=[]):
        if property_names:
            namespaces = []
            namespace_properties = {}
            for item in property_names:
                if '.' not in item:
                    item = "core." + item
                Identifier(item, self._namespace_schemas)  # check if the identifier is valid
                namespace, identifier = item.split('.')
                if namespace not in namespaces:
                    namespaces.append(namespace)
                    namespace_properties[namespace] = ['uuid']  # always add uuid to determine if namespace exists
                if identifier != 'uuid':
                    namespace_properties[namespace].append(identifier)
            join_set = set(namespaces)
            description = [(namespace, namespace_properties[namespace]) for namespace in namespaces]
        else:
            join_set = set(namespaces)
            description = [("core", list(self._namespace_schema("core")))]
            for namespace in join_set:
                description.append((namespace, ["uuid"] + list(self._namespace_schema(namespace))))

        # Parse the where clause.
        where_clause, where_parameters = "", {}
        if where:
            ast = parse_and_analyze(where, self._namespace_schemas, parameters)
            visitor = _WhereExpressionVisitor(self._rewriter_table, self._column_name, self._named_placeholder)
            where_expr, where_parameters, where_namespaces = visitor.do_visit(ast)
            if where_expr:
                join_set.update(where_namespaces)
                where_clause = "WHERE %s" % where_expr

        # Parse the order by clause.
        order_by_clause = ""
        if order_by:
            order_by_list, order_by_namespaces = self._build_order_by_list(order_by)
            if order_by_list:
                join_set.update(order_by_namespaces)
                order_by_clause = "ORDER BY %s" % ", ".join(order_by_list)

        # Parse the limit clause.
        limit_clause = ""
        if limit is not None:
            try:
                limit = int(limit)
            except (TypeError, ValueError):
                raise Error("limit %r must be a positive integer" % limit)

            if limit < 0:
                raise Error("limit %r must be a positive integer" % limit)

            limit_clause = "LIMIT %d" % limit

        # Generate the SELECT clause.
        select_list = []
        for namespace, identifiers in description:
            select_list.extend([self._column_name(namespace, identifier) for identifier in identifiers])
        select_clause = "SELECT %s" % ", ".join(select_list)

        # Generate the FROM clause.
        from_clause = "FROM %s" % self._table_name("core")

        join_set.discard("core")
        for namespace in join_set:
            from_clause = "%s LEFT JOIN %s USING (uuid)" % (from_clause, self._table_name(namespace))

        # Generate the complete query.
        query = "%s %s" % (select_clause, from_clause)
        if where_clause:
            query = "%s %s" % (query, where_clause)
        if order_by_clause:
            query = "%s %s" % (query, order_by_clause)
        if limit_clause:
            query = "%s %s" % (query, limit_clause)

        return query, where_parameters, description

    def _build_order_by_list(self, order_by_items):
        order_by_list, namespaces = [], set()
        for item in order_by_items:
            direction = "DESC" if item.startswith("-") else "ASC"
            name = item[1:] if item.startswith("+") or item.startswith("-") else item

            segments = name.split('.')
            if len(segments) == 1:
                namespace, name = 'core', segments[0]
            elif len(segments) == 2:
                namespace, name = name.split(".")
            else:
                raise Error("invalid property name: %r" % name)

            if name not in self._namespace_schema(namespace):
                raise Error("no property: %r defined within namespace: %r" % (name, namespace))

            namespaces.add(namespace)
            order_by_list.append(self._column_name(namespace, name) + " " + direction)

        return order_by_list, namespaces

    def _column_name(self, namespace, identifier):
        return self._table_name(namespace) + "." + identifier

    def _namespace_schema(self, namespace):
        try:
            return self._namespace_schemas[namespace]
        except KeyError:
            raise Error("undefined namespace: \"%s\"" % namespace)

    def _type(self, type):
        try:
            return self._type_map[type]
        except KeyError:
            raise Error("type not supported by backend: %r" % type.name())



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import codecs
import struct

from muninn.enum import Enum
from muninn.exceptions import Error
from muninn.geometry import LinearRing, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon
from muninn.visitor import Visitor


class GeometryType(Enum):
    _items = ("GEOMETRY", "POINT", "LINESTRING", "POLYGON", "MULTIPOINT", "MULTILINESTRING", "MULTIPOLYGON")


class EWKBEncoder(Visitor):
    def __init__(self, little_endian=True, srid=4326):
        self.endianness = int(little_endian)
        self.srid = srid
        self.prefix = (">", "<")[little_endian]

    def visit(self, visitable, tagged=True, srid=True):
        return super(EWKBEncoder, self).visit(visitable, tagged, srid)

    def visit_Point(self, visitable, tagged, srid):
        ewkb = self._encode("dd", visitable.x, visitable.y)
        return self._encode_tag(GeometryType.POINT, srid) + ewkb if tagged else ewkb

    def visit_LineString(self, visitable, tagged, srid):
        ewkb = self._encode("I", len(visitable))
        ewkb += b"".join([self.visit(point, False) for point in visitable])
        return self._encode_tag(GeometryType.LINESTRING, srid) + ewkb if tagged else ewkb

    def visit_LinearRing(self, visitable, tagged, srid):
        if len(visitable) == 0:
            ewkb = self._encode("I", 0)
        else:
            ewkb = self._encode("I", len(visitable))
            ewkb += b"".join([self.visit(point, False) for point in visitable])
        return self._encode_tag(GeometryType.LINESTRING, srid) + ewkb if tagged else ewkb

    def visit_Polygon(self, visitable, tagged, srid):
        ewkb = self._encode("I", len(visitable))
        ewkb += b"".join([self.visit(ring, False) for ring in visitable])
        return self._encode_tag(GeometryType.POLYGON, srid) + ewkb if tagged else ewkb

    def visit_MultiPoint(self, visitable, tagged, srid):
        ewkb = self._encode("I", len(visitable))
        ewkb += b"".join([self.visit(point, True, False) for point in visitable])
        return self._encode_tag(GeometryType.MULTIPOINT, srid) + ewkb if tagged else ewkb

    def visit_MultiLineString(self, visitable, tagged, srid):
        ewkb = self._encode("I", len(visitable))
        ewkb += b"".join([self.visit(line_string, True, False) for line_string in visitable])
        return self._encode_tag(GeometryType.MULTILINESTRING, srid) + ewkb if tagged else ewkb

    def visit_MultiPolygon(self, visitable, tagged, srid):
        ewkb = self._encode("I", len(visitable))
        ewkb += b"".join([self.visit(polygon, True, False) for polygon in visitable])
        return self._encode_tag(GeometryType.MULTIPOLYGON, srid) + ewkb if tagged else ewkb

    def default(self, visitable, tagged, srid):
        raise Error("unsupported type: %s" % type(visitable).__name__)

    def _encode_tag(self, geometry_type, srid):
        if srid and self.srid is not None:
            return self._encode("BII", self.endianness, geometry_type | 0x20000000, self.srid)
        return self._encode("BI", self.endianness, geometry_type)

    def _encode(self, format, *args):
        try:
            return struct.pack(self.prefix + format, *args)
        except struct.error as _error:
            raise Error("encoding error: %s" % str(_error))


class EWKBStream(object):
    def __init__(self, ewkb):
        self.ewkb = ewkb
        self.offset = 0
        self.prefix = (">", "<")[self._decode("=", "B")]

    def decode(self, format):
        return self._decode(self.prefix, format)

    def tail(self):
        return self.ewkb[self.offset:]

    def _decode(self, prefix, format):
        format = prefix + format

        try:
            start, end = self.offset, self.offset + struct.calcsize(format)
            values = struct.unpack(format, self.ewkb[start:end])
        except struct.error as _error:
            raise Error("decoding error: %s" % str(_error))

        self.offset = end
        return values[0] if len(values) == 1 else values


def _decode_point(stream):
    return Point(*stream.decode("dd"))


def _decode_line_string(stream):
    count = stream.decode("I")
    return LineString([_decode_point(stream) for _ in range(count)])


def _decode_linear_ring(stream):
    count = stream.decode("I")
    if count == 0:
        return LinearRing()

    if count < 4:
        raise Error("linear ring should be empty or should contain >= 4 points")

    points = [_decode_point(stream) for _ in range(count)]
    if points[-1] != points[0]:
        raise Error("linear ring should be closed")

    return LinearRing(points)


def _decode_polygon(stream):
    count = stream.decode("I")
    return Polygon([_decode_linear_ring(stream) for _ in range(count)])


def _decode_geometry_sequence(stream, expected_ewkb_type):
    sequence, count = [], stream.decode("I")
    for _ in range(count):
        stream = EWKBStream(stream.tail())
        sequence.append(_decode_ewkb(stream, expected_ewkb_type))
    return sequence


def _decode_multi_point(stream):
    return MultiPoint(_decode_geometry_sequence(stream, GeometryType.POINT))


def _decode_multi_line_string(stream):
    return MultiLineString(_decode_geometry_sequence(stream, GeometryType.LINESTRING))


def _decode_multi_polygon(stream):
    return MultiPolygon(_decode_geometry_sequence(stream, GeometryType.POLYGON))


def _decode_ewkb(stream, expected_ewkb_type=None):
    ewkb_type = stream.decode("I")
    ewkb_type, ewkb_flags = ewkb_type & 0x00FFFFFF, ewkb_type >> 28

    if expected_ewkb_type is not None and ewkb_type != expected_ewkb_type:
        raise Error("unexpected EWKB type code: %s (expected: %s)" % (ewkb_type, expected_ewkb_type))

    if ewkb_flags == 0x02:
        srid = stream.decode("I")
        if srid != 4326:
            raise Error("unsupported SRID code: %d" % srid)
    elif ewkb_flags != 0x00:
        raise Error("unsupported EWKB type flags: %d" % ewkb_flags)

    if ewkb_type == GeometryType.POINT:
        return _decode_point(stream)
    elif ewkb_type == GeometryType.LINESTRING:
        return _decode_line_string(stream)
    elif ewkb_type == GeometryType.POLYGON:
        return _decode_polygon(stream)
    elif ewkb_type == GeometryType.MULTIPOINT:
        return _decode_multi_point(stream)
    elif ewkb_type == GeometryType.MULTILINESTRING:
        return _decode_multi_line_string(stream)
    elif ewkb_type == GeometryType.MULTIPOLYGON:
        return _decode_multi_polygon(stream)

    raise Error("unsupported EWKB type code: %d" % ewkb_type)


def encode_ewkb(geometry):
    return EWKBEncoder().visit(geometry)


def encode_hexewkb(geometry):
    return codecs.encode(encode_ewkb(geometry), "hex").decode().upper()


def decode_ewkb(ewkb):
    return _decode_ewkb(EWKBStream(ewkb))


def decode_hexewkb(hexewkb):
    return decode_ewkb(codecs.decode(hexewkb, "hex"))



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import datetime
import json
import uuid
from muninn._compat import long
from muninn._compat import string_types as basestring

from muninn import geometry


def prefix_message_with_path(path, message):
    if not path:
        return message
    if path.endswith(":"):
        return path + " " + message
    return path + ": " + message


def join(path, *args):
    if not path:
        return ".".join(args)
    if path.endswith(":"):
        return path + " " + ".".join(args)
    return path + "." + ".".join(args)


class optional(object):
    def __init__(self, type):
        self.type = type


class Type(object):
    def __init__(self, optional=False, index=False):
        self.optional = optional
        self.index = index

    @classmethod
    def name(cls):
        return getattr(cls, "_alias", cls.__name__.lower())


class Long(Type):
    @classmethod
    def validate(cls, value):
        if type(value) is not int and type(value) is not long or value < -9223372036854775808 \
                or value > 9223372036854775807:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Integer(Type):
    @classmethod
    def validate(cls, value):
        if type(value) is not int and type(value) is not long or value < -2147483648 or value > 2147483647:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Real(Type):
    @classmethod
    def validate(cls, value):
        if type(value) is not float:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Boolean(Type):
    @classmethod
    def validate(cls, value):
        if type(value) is not bool:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Text(Type):
    @classmethod
    def validate(cls, value):
        if not isinstance(value, basestring):
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Namespace(Type):
    @classmethod
    def validate(cls, value):
        if not isinstance(value, basestring):
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Timestamp(Type):
    @classmethod
    def validate(cls, value):
        if type(value) is not datetime.datetime:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class JSON(Type):
    @classmethod
    def validate(cls, value):
        try:
            json.dumps(value)
        except Exception:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class UUID(Type):
    @classmethod
    def validate(cls, value):
        if type(value) is not uuid.UUID:
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Geometry(Type):
    @classmethod
    def validate(cls, value):
        if not isinstance(value, geometry.Geometry):
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Container(Type):
    pass


class Sequence(Container):
    sub_type = None

    @classmethod
    def validate(cls, value, partial=False, path=""):
        path = "%s:" % cls.name() if not path else path

        try:
            iterator = iter(value)
        except TypeError:
            raise ValueError(prefix_message_with_path(path, "expected a sequence"))

        for index, sub_value in enumerate(iterator):
            if issubclass(cls.sub_type, Container):
                cls.sub_type.validate(sub_value, partial, path + "[%d]" % index)
            else:
                try:
                    cls.sub_type.validate(sub_value)
                except ValueError as _error:
                    raise ValueError(prefix_message_with_path(path + "[%d]" % index, str(_error)))


class MetaMapping(type):
    def __new__(meta, name, bases, dct):
        class_dct = {}
        items = {}
        for key, value in dct.items():
            # class: not optional, no index
            if isinstance(value, type) and issubclass(value, Type):
                items[key] = (value, False, False)

            # instance: get optional, index from instance
            elif isinstance(value, Type):
                items[key] = (type(value), value.optional, value.index)

            # optional() wrapper: optional, no index
            elif type(value) is optional:
                items[key] = (value.type, True, False)
            else:
                class_dct[key] = value

        assert "_items" not in class_dct
        class_dct["_items"] = items

        return super(MetaMapping, meta).__new__(meta, name, bases, class_dct)

    def validate(cls, value, partial=False, path=""):
        path = "%s:" % cls.name() if not path else path

        validated = 0
        for sub_name, (sub_type, sub_optional, sub_index) in cls._items.items():
            try:
                sub_value = value[sub_name]
            except TypeError:
                raise ValueError(prefix_message_with_path(path, "expected a mapping"))
            except KeyError:
                if not partial and not sub_optional:
                    raise ValueError(prefix_message_with_path(join(path, sub_name), "no value for mandatory item"))
            else:
                if not sub_optional or sub_value is not None:
                    if issubclass(sub_type, Container):
                        sub_type.validate(sub_value, partial, join(path, sub_name))
                    else:
                        try:
                            sub_type.validate(sub_value)
                        except ValueError as _error:
                            raise ValueError(prefix_message_with_path(join(path, sub_name), str(_error)))

                validated += 1

        if validated != len(value):
            extra_names = set(value) - set(cls._items)
            raise ValueError(prefix_message_with_path(path, "undefined item: %r" % extra_names.pop()))

    # TODO store more nicely (at least not positionally)
    def __getitem__(cls, name):
        return cls._items[name][0]

    def is_optional(cls, name):
        return cls._items[name][1]

    def has_index(cls, name):
        return cls._items[name][2]

    def __len__(cls):
        return len(cls._items)

    def __iter__(cls):
        return iter(cls._items)

    def __contains__(cls, name):
        return name in cls._items


Mapping = MetaMapping('Mapping', (Container,), {})



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import errno
import hashlib
import os
import shutil
import tempfile

from muninn._compat import string_types as basestring
from muninn._compat import path_utf8, encode, decode


class TemporaryDirectory(object):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def __enter__(self):
        self._path = tempfile.mkdtemp(*self._args, **self._kwargs)
        return self._path

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self._path)
        return False


def split_path(path):
    """Recursively apply os.path.split() to split a path into a sequence of path segments.

    Example:
        [segment for segment in split_path("/a/b/c/d.txt")] returns ["/", "a", "b", "c", "d.txt"]

    """
    head, tail = os.path.split(path)
    if not tail:
        head, tail = os.path.split(head)

    if head.endswith(os.sep) or head.endswith('/'):  # return e.g. '/' or 'c:/'
        yield head
    elif head:
        for segment in split_path(head):
            yield segment

    if tail:
        yield tail


def is_sub_path(sub_path, path, allow_equal=False):
    """Determine if a path is a sub path of (is contained in) another path.

    The 'allow_equal' flag determines whether paths that are equal are considered sub paths of eachother or not.

    Both paths are split into separate segments using os.split(), and compared segment by segment. This avoids the
    problem where "/a/bb/c" is considered a sub path of "/a/b", as would happen when comparing using str.startswith()
    or os.path.commonprefix().

    """
    sub_path_segments = [segment for segment in split_path(sub_path)]
    path_segments = [segment for segment in split_path(path)]

    if allow_equal:
        if len(sub_path_segments) < len(path_segments):
            return False
    else:
        if len(sub_path_segments) <= len(path_segments):
            return False

    for sub_path_segment, path_segment in zip(sub_path_segments, path_segments):
        if sub_path_segment != path_segment:
            return False

    return True


def make_path(path, mode=0o777):
    """Try to create the specified path, creating parent directories where needed. If the path already exists, this is
    _not_ considered an error. This is similar to "mkdir -p" and in constrast to os.makedirs(). The latter raises an
    exception if the leaf directory exists.

    Keyword arguments:
    mode -- On some systems, mode is ignored. Where it is used, the current umask value is first masked out.
            The default mode is 0777 (octal). See also the documentation of os.mkdir().

    """
    try:
        os.makedirs(path)
    except EnvironmentError as _error:
        # If the leaf directory cannot be created because an entity with the same name already exists, do not consider
        # this an error _if_ this entity is (a symbolic link to) a directory, instead of e.g. a file.
        if _error.errno != errno.EEXIST or not os.path.isdir(path):
            raise


def copy_path(source, target, resolve_root=False, resolve_links=False):
    """Recursively copy the source path to the destination path. The destination path should not exist. Directories are
    copied as (newly created) directories with the same names, files are copied by copying their contents
    (using shutil.copy()).

    Keyword arguments:
    resolve_root -- If set to True and if the top-level file/directory for the source tree is a symbolic link then the
                    source tree is copied by copying the linked target to the destination tree.
                    If set to False, the top-level link in the source tree is copied based on the setting of
                    the resolve_links parameter.

    resolve_links -- If set to True, all symbolic links in the source tree are copied by recursively copying the linked
                     targets to the destination tree. Note that this could lead to infinite recursion.
                     If set to False, symbolic links in the source tree are copied as (newly created) symbolic links
                     in the destination tree.

    """
    def _copy_path_rec(source, target, resolve_root, resolve_links):
        assert os.path.exists(source) or os.path.islink(source)

        # Refuse to copy into a dangling symlink.
        if os.path.islink(target) and not os.path.exists(target):
            raise IOError("target is a dangling symlink: %s" % target)

        if os.path.islink(source) and not (resolve_links or resolve_root):
            if os.path.exists(target):
                # This will fail if target is a directory, which is the intended behaviour.
                os.remove(target)

            # If target is a dangling symlink, os.path.exists() returns False and the creation of the symlink below
            # will fail, which is the intended behaviour.
            os.symlink(os.readlink(source), target)

        elif os.path.isdir(source):
            if not os.path.exists(target):
                # This will fail if target exists, which is the intended behaviour. Note that if the target is a
                # dangling symlink, the creation of the directory below will fail as well, which is intended.
                os.mkdir(target)

            for basename in os.listdir(source):
                source_path = os.path.join(source, basename)
                target_path = os.path.join(target, basename)
                # The resolve_root option should only have an effect during the initial call to _copy_path_rec().
                _copy_path_rec(source_path, target_path, False, resolve_links)

        else:
            shutil.copyfile(source, target)
            shutil.copystat(source, target)

    # If the source ends in a path separator and it is a symlink to a directory, then the symlink will be resolved even
    # if resolve_root is set to False. Disallow a root that refers to a file and has a trailing path separator.
    if source.endswith(os.path.sep):
        if not os.path.isdir(os.path.dirname(source)):
            raise IOError("not a directory: %s" % source)
        else:
            resolve_root = True

    # If the target is a directory, copy the source _into_ the target, unless the source path ends in a path separator.
    if os.path.isdir(target) and not source.endswith(os.path.sep):
        target = os.path.join(target, os.path.basename(source))

    # Perform the recursive copy.
    _copy_path_rec(source, target, resolve_root, resolve_links)


def remove_path(path):
    if not os.path.isdir(path) or os.path.islink(path):
        os.remove(path)
    else:
        shutil.rmtree(path)


def hash_string(string, hash_func):
    hash = hash_func()
    hash.update(string)
    return encode(hash.hexdigest())


def hash_file(path, block_size, hash_func):
    hash = hash_func()
    with open(path, "rb") as stream:
        while True:
            # Read a block of character data.
            data = stream.read(block_size)
            if not data:
                return encode(hash.hexdigest())

            # Update hash.
            hash.update(data)


# NB. os.path.islink() can be True even if neither os.path.isdir() nor os.path.isfile() is True.
# NB. os.path.exists() is False for a dangling symbolic link, even if the symbolic link itself does exist.
def product_hash(roots, resolve_root=True, resolve_links=False, force_encapsulation=False,
                 block_size=65536, hash_type=None):
    hash_func = getattr(hashlib, hash_type or 'sha1')

    def _product_hash_rec(root, resolve_root, resolve_links, hash_func, block_size):
        if os.path.islink(root) and not (resolve_root or resolve_links):
            # Hash link _contents_.
            return hash_string(path_utf8(os.readlink(root)), hash_func)

        elif os.path.isfile(root):
            # Hash file contents.
            return hash_file(root, block_size, hash_func)

        elif os.path.isdir(root):
            # Create a fingerprint of the directory by computing the hash of (for each entry in the directory) the hash
            # of the entry name, the type of entry (link, file, or directory), and the hash of the contents of the
            # entry.
            hash = hash_func()
            for basename in sorted(os.listdir(root)):
                hash.update(hash_string(path_utf8(basename), hash_func))

                path = os.path.join(root, basename)
                if os.path.islink(path) and not (resolve_root or resolve_links):
                    hash.update(b"l")
                elif os.path.isdir(path):
                    hash.update(b"d")
                else:
                    hash.update(b"f")

                hash.update(_product_hash_rec(path, False, resolve_links, hash_func, block_size))

            return encode(hash.hexdigest())

        else:
            raise IOError("path does not refer to a regular file or directory: %s" % root)

    if isinstance(roots, basestring):
        roots = [roots]

    if len(roots) == 1 and not force_encapsulation:
        return hash_type + ':' + decode(_product_hash_rec(roots[0], resolve_root, resolve_links, hash_func, block_size))

    hash = hash_func()
    for root in sorted(roots):
        hash.update(hash_string(path_utf8(os.path.basename(root)), hash_func))

        if os.path.islink(root) and not (resolve_root or resolve_links):
            hash.update(b"l")
        elif os.path.isdir(root):
            hash.update(b"d")
        else:
            hash.update(b"f")

        hash.update(_product_hash_rec(root, resolve_root, resolve_links, hash_func, block_size))

    return hash_type + ':' + hash.hexdigest()


def product_size(roots, resolve_root=True, resolve_links=False):
    def _product_size_rec(root, resolve_root, resolve_links):
        if os.path.islink(root) and not (resolve_root or resolve_links):
            # Use size of the symbolic link itself (instead of the size of its target).
            return os.lstat(root).st_size

        elif os.path.isfile(root):
            return os.stat(root).st_size

        elif os.path.isdir(root):
            return sum([_product_size_rec(os.path.join(root, base), False, resolve_links) for base in os.listdir(root)])

        else:
            raise IOError("path does not refer to a regular file or directory: %s" % root)

    if isinstance(roots, basestring):
        roots = [roots]

    return sum([_product_size_rec(root, resolve_root, resolve_links) for root in roots])


def quoted_list(lst, quote_text='"', join_text=", "):
    '''returns a string where all items are surrounded by quotes and joined'''
    return join_text.join([quote_text + str(x) + quote_text for x in lst])


def fwd_join(*args):
    return '/'.join([a for a in args if a != ''])



from setuptools import setup
import sys

pyversion = sys.version_info
if pyversion[0] == 2 and pyversion[1] >= 7:
    python_req = ">=2.7"
elif pyversion >= (3, 6):
    python_req = ">=3.6"
else:
    # python_requires is only availabe since setuptools 24.2.0 and pip 9.0.0
    sys.exit("Python 2.7 (or newer) or 3.6 (or newer) is required to use this package.")


requirements = []

setup(
    name="muninn",
    version="7.2.1",
    description="Data product catalogue and archive system",
    url="https://github.com/stcorp/muninn",
    author="S[&]T",
    license="BSD",
    packages=["muninn", "muninn.tools", "muninn.database", "muninn.storage"],
    entry_points={"console_scripts": [
        "muninn-attach = muninn.tools.attach:main",
        "muninn-destroy = muninn.tools.destroy:main",
        "muninn-export = muninn.tools.export:main",
        "muninn-hash = muninn.tools.hash:main",
        "muninn-info = muninn.tools.info:main",
        "muninn-ingest = muninn.tools.ingest:main",
        "muninn-list-tags = muninn.tools.list_tags:main",
        "muninn-prepare = muninn.tools.prepare:main",
        "muninn-pull = muninn.tools.pull:main",
        "muninn-remove = muninn.tools.remove:main",
        "muninn-retrieve = muninn.tools.retrieve:main",
        "muninn-search = muninn.tools.search:main",
        "muninn-strip = muninn.tools.strip:main",
        "muninn-summary = muninn.tools.summary:main",
        "muninn-sync = muninn.tools.sync:main",
        "muninn-tag = muninn.tools.tag:main",
        "muninn-untag = muninn.tools.untag:main",
        "muninn-update = muninn.tools.update:main",
    ]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    python_requires=python_req,
    install_requires=requirements
)

# For the postgres backend you will need psycopg 2.2 or higher.

# For the sqlite backend you will need sqlite/pysqlite2 with the mod_spatialite
# extension.

# For muninn-pull using http requests you will need 'requests' 2.13.0 or higher
# For muninn-pull using sftp requests you will need 'paramiko'
# For muninn-pull with oauth2 authentication you weel need 'requests-oauthlib'
# and 'oauthlib'

# To get more output formatting options for muninn-search install 'tabulate'

# To see progress bars for muninn-update install 'tqdm'



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import argparse
import datetime

import muninn

from muninn.tools.utils import create_parser, parse_args_and_run


def valid_date(arg: str) -> datetime:
    for format_string in ("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S.%f"):
        try:
            return datetime.datetime.strptime(arg, format_string)
        except ValueError:
            pass
    raise argparse.ArgumentTypeError(f"invalid datetime: {arg} (expected 'YYYY-MM-DD(THH:MM:SS(.ffffff))')")


def run(args):
    with muninn.open(args.archive) as archive:
        archive.sync(args.synchronizer, args.product_types, args.start, args.end, args.force)


def main():
    parser = create_parser(description="Synchronize a muninn archive")
    parser.add_argument("archive", metavar="ARCHIVE", help="archive identifier")
    parser.add_argument("synchronizer", help="name of the synchronizer to use")
    parser.add_argument("-p", "--product-types", nargs="*",
                        help="the product types to be synchronized")
    parser.add_argument('--start', type=valid_date, help='start datetime of synchronization range')
    parser.add_argument('--end', type=valid_date, help='end datetime of synchronization range')
    parser.add_argument("-f", "--force", action="store_true", default=False,
                        help="force update, even if entries have not changed")
    return parse_args_and_run(parser, run)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2025 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function


class MetaEnum(type):
    def __new__(meta, name, bases, dct):
        class_dct = dct.copy()
        for value, item in enumerate(dct.get('_items', [])):
            class_dct[item] = value
        return super(MetaEnum, meta).__new__(meta, name, bases, class_dct)

    def count(cls):
        return len(cls._items)

    def valid(cls, value):
        return value >= 0 and value < len(cls._items)

    def items(cls):
        return cls._items

    def to_string(cls, value):
        if not cls.valid(value):
            raise ValueError("enumeration: %s has no item with value: %d" % (cls.__name__, value))
        return cls._items[value]

    def from_string(cls, value):
        try:
            return cls._items.index(value)
        except ValueError:
            raise ValueError("enumeration: %s does not contain: %s" % (cls.__name__, value))


Enum = MetaEnum('Enum', (), {})
